################################################################################
# 3hauka.tcl
# Copyright (c) 2001-2004 hex@faerion.oss
#
#        .
#  : https://he3hauka.sourceforge.net/
#
# $Id: 3hauka.tcl,v 1.3 2004/04/14 09:53:06 vhex Exp $
#
################################################################################

namespace eval tpub {
	#   
	array unset conf
	array unset channels

	#   
	set conf(fnquestions) "[file dirname [info script]]/3hauka.txt"

	#      
	set conf(minanswers) 5

	# ,      
	#   ,         ( )
	set conf(maxidle) [expr 60 * 60 * 24 * 31] ;#  

	#   , : {8-12 16-18}
	set conf(timelimit) {0-24}

	#    
	set conf(url) "http://he3hauka.sourceforge.net"
	# set conf(url-top) "http://quiz.forestnet.org/top.var"
	# set conf(url-cmd) "http://quiz.forestnet.org/#commands"

	# ,       ,
	#      
	set conf(scoreboard) "[file dirname [info script]]/3hauka.score"

	#   
	if {[file exists [set _fname [file dirname [info script]]/3hauka.conf]]} {
		source $_fname
	}
}

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

###      
# proc ::tpub::tolower {text} { return [string map { } [string tolower $text]] }
proc ::tpub::tolower {text} { return [string map {     } [string tolower $text]] }

proc ::tpub::toupper {text} { return [string map {     } [string toupper $text]] }

###     (c) s7ream
proc ::tpub::numgrp {number} { switch -glob -- "$number" { *11 {return 3} *12 {return 3} *13 {return 3} *14 {return 3} *1 {return 1} *2 {return 2} *3 {return 2} *4 {return 2} default {return 3} } }

###    
proc ::tpub::playground {command name {option undefined}} {
	variable channels

	switch -exact $command {
		"exists" {
			return [info exists channels([string tolower $name])]
		}
		"check" {
			if {![info exists channels([string tolower $name])]} {
				putlog "he3hauka.tcl: playground: bad channel $name"
				return 0
			}
			return [expr { [lsearch -exact [set channels([string tolower $name])] $option] == -1 ? 0 : 1 }]
		}
	}
}

proc ::tpub::maskhost {uhost} {
	return [::maskhost $uhost]
#	regsub {(.+@)[^.]+(\..+)} $uhost {\1*\2} host
#	return $host
}

proc ::tpub::getversion {} { variable conf
	return "1.10"
}

#   
proc putout {cmd dest msg} { global botnick
	set maxlen [expr 510 - [string length "$botnick![getchanhost $botnick] $cmd $dest :"]]
	while {$msg != ""} {
		set text [expr {[string length $msg] > $maxlen ? [string range $msg 0 [string last \040 [string range $msg 0 $maxlen]]] : $msg}]
		putserv "$cmd $dest :$text"
		set msg [string trim [string range $msg [string length $text] end]]
	}
}

proc putmsg {nick msg} { putout PRIVMSG $nick $msg }
proc putnot {nick msg} { putout NOTICE $nick $msg }
proc putact {dest msg} { putserv "PRIVMSG $dest :\001ACTION $msg\001" }
proc putpvt {nick msg} { puthelp "PRIVMSG $nick :$msg" }

###      
proc ::tpub::usage {nick cmd {arg ""} {info ""}} {
	set not "\002$cmd\002"
	if {$arg != ""} { append not " $arg" }
	if {$info != ""} { append not " - $info"}

	set aliases ""
	foreach item [lsort [info commands ::tpub::cmd_*]] {
		if {[string match "return \\\[cmd_$cmd * * *\\\]" [string trim [info body $item]]]} { append aliases " [string range $item 12 end]" }
	}
	if {$aliases != ""} {
		append not ". :$aliases"
	}

	putnot $nick ": $not"
	return 0
}


#      
proc ::tpub::checktime {} { variable conf
	if {![info exists conf(timelimit)]} { return 1 }

	foreach item [split $conf(timelimit)] {
		#   ,     
		if {[string is int $item] && [expr int($item) == int([clock format [clock seconds] -format %H])]} {
			return 1
		}

		#     
		if {![regexp {^(\d+(:\d+)?)?-(\d+(:\d+)?)?.*$} $item t1 t1 t2 t2]} {
			putlog "he3hauka: wrong time interval: $item"
			continue
		}

		#    
		if {$t1 == ""} { set t1 "00:00" }
		if {$t2 == ""} { set t2 "24:00" }

		#   
		if {[string is int $t1]} { append t1 ":00" }
		if {[string is int $t2]} { append t2 ":00" }

		#    ..
		set t1 [expr double([string map {: .} $t1])]
		set t2 [expr double([string map {: .} $t2])]

		#  
		set t [expr double([clock format [clock seconds] -format "%H.%M"]) ]

		# ,   
		if {$t >= $t1 && $t <= $t2} { return 1 }
	}

	#   ..
	return 0
}

#   
# question rehash -    ,   
# question get -      ( 
#                 rehash,   ),  :
#                {question {answer answer answer}}
proc ::tpub::question {cmd {text ""}} { variable conf; variable qpos
	switch -exact -- [string tolower $cmd] {
		"count" {
			if {![info exists qpos]} { return 0 }
			return [llength $qpos]
		}

		"rehash" {
			if {[info exists qpos]} { unset qpos }
			set qpos {}
			set ::tpub::qcount 0
			if {[catch {open $conf(fnquestions) r} fileID]} { return 0 }
			while {![eof $fileID]} {
				set pos [tell $fileID]
				incr ::tpub::qcount
				if {[set line [string trim [gets $fileID]]] == ""} { continue }
				if {[string range $line 0 0] == "#"} { continue }
				lappend qpos $pos
			}
			close $fileID
			return [llength $qpos]
		}

		"get" {
			# random question
			if {![info exists qpos] || ![llength $qpos]} {
				putlog "Out of questions, rehashing..."
				if {![question rehash]} {
					putlog "No questions available, eek."
					return {}
				}
			}
			set idx [expr int(rand() * [llength $qpos])]
			set fpos [lindex $qpos $idx]
			set qpos [lreplace $qpos $idx $idx]

			if {[catch {open $conf(fnquestions) r} fileID]} { return {} }
			seek $fileID $fpos start
			set line [gets $fileID]
			set line [split [string trim $line] |]
			set line "[toupper [string range $line 0 0]][string range $line 1 end]"
			close $fileID

			return [list [lindex $line 0] [split [lrange $line 1 end] ,]]
		}

		"find" {
			if {$text == ""} {
				error "wrong # args: should be \"questions find pattern\""
			}
			set idx -1
			if {[catch {open $conf(fnquestions) r} fileID]} { return $idx }
			while {![eof $fileID]} {
				if {[string range [set line [gets $fileID]] 0 0] == "#"} { continue }
				incr idx
				if {[lindex [split $line |] 0] == $text} {
					close $fileID
					return $idx
				}
			}
			close $fileID
			return -1
		}

		"del" {
			if {$text == ""} {
				error "wrong # args: should be \"questions del number\""
			}
			if {![string is integer $text]} {
				error "bad option \"$text\": must be an integer"
			}
			if {[catch {open $conf(fnquestions) r+} fileID]} { return 0 }
			while {![eof $fileID]} {
				set pos [tell $fileID]
				if {[string range [gets $fileID] 0 0] == "#"} { continue }
				if {![expr int($text)]} {
					puts -nonewline $fileID "#"
					close $fileID
					return 1
				}
				incr text -1
			}
			close $fileID
		}

		"add" {
			if {$text == ""} {
				error "" "wrong # args: should be \"questions add question test|answer|..\""
			}
			if {[string first | $text] == -1} {
				error "" "wrong question format, must be: \002question text|answer\002[|answer|...]"
			}
			if {[catch {open $conf(fnquestions) a} fileID]} {
				error "" "could not open questions database for writing."
			}
			puts $fileID " [string trim $text]"
			close $fileID
		}

		default {
			error "bad option \"$cmd\": must be rehash, get, find or del"
		}
	}
	return 0
}


#      
proc ::tpub::get_new_host {host} { variable conf
	if {![info exists conf(static_hosts)]} {
		return "*!$host"
	}

	foreach item $conf(static_hosts) {
		if {[string match $item $host]} {
			return "*!$host"
		}
	}

	return [maskhost $host]
}


#   
proc ::tpub::users {cmd {text {}}} {
	variable conf

	switch -exact -- $cmd {
		"purge" {
			set killed 0
			foreach hand [userlist] {
				if {[set xtra [getuser $hand xtra quiz]] == ""} { continue }

				# check if one has a channel record
				if {[chattr $hand] != "-"} { continue }
				foreach chan [array names channels] {
					if {[haschanrec $hand $chan]} { continue }
				}

				array set usr $xtra

				# linked nick, won't remove anyways
				if {[info exists usr(link)]} { continue }

				if {(![info exists usr(answers)] || $usr(answers) < $conf(minanswers)) || ([info exists usr(last)] && [expr [clock seconds] - $usr(last)] > $conf(maxidle))} {
					incr killed
					deluser $hand
				}
			}
			return $killed
		}

		"list" {
			# list valid users
			set users {}
			foreach hand [userlist] {
				if {[set xtra [getuser $hand xtra quiz]] == ""} { continue }
				array unset usr
				array set usr $xtra
				if {[info exists usr(halt)]} { continue }
				if {[info exists usr(last)] && [expr [clock seconds] - $usr(last)] < 604800 && ![info exists usr(link)]} {
					lappend users [list $hand $usr(score) [format %d.%08d $usr(score) [expr 999999 - $usr(answers)]]]
				}
			}

			# sort by place and score
			set sorted {}
			foreach item [lsort -real -decreasing -index 2 $users] {
				lappend sorted [lrange $item 0 1]
			}
			return $sorted
		}

		"exists" {
			return [validuser $text]
		}

		"isintop" {
			if {$text == ""} {
				error "" "wrong # args: should be \"users isintop user\""
			}
			foreach item [lrange [users list] 0 4] {
				if {![string compare -nocase $text [lindex $item 0]]} { return 1 }
			}
			return 0
		}

		"get" {
			if {$text == ""} {
				error "wrong # args: should be \"users get nick\""
			}
			if {![validuser $text]} {
				error "no such user - $text"
			}
			return [getuser $text xtra quiz]
		}

		"set" {
			if {[validuser [set hand [lindex $text 0]]]} {
				setuser $hand xtra quiz [join [lrange $text 1 end]]
				return 1
			} else {
				return 0
			}
		}

		"check" {
			if {$text == ""} {
				error "wrong # args: should be \"users check hand\""
			}

			set hand [lindex $text 0]

			if {![validuser $hand]} {
				error "users check: invalid handle: $hand"
				return
			}

			#    
			array set data [getuser $hand xtra quiz]
			if {![info exists data(score)]} { set data(score) 0 }
			if {![info exists data(cont)]} { set data(cont) 0 }
			if {![info exists data(quick)]} { set data(quick) 0 }
			if {![info exists data(answers)]} { set data(answers) 0 }
			if {![info exists data(last)]} { set data(last) [clock seconds] }
			setuser $hand xtra quiz [array get data]
		}

		default {
			error "bad option \"$cmd\": must be purge, load, save, list, get or set"
		}
	}
	return 0
}


###    
proc ::tpub::game_time {} { variable game
	if {![info exists game] || ![info exists game(started)]} { return " " }

	set taken [expr ([clock seconds] - $game(started)) / 60]
	set min [expr $taken % 60]
	set hrs [expr ($taken - ($taken % 60)) / 60]
	if {!$taken} { return " " }

	set msg ""
	if {$hrs} { append msg " $hrs [lindex {.   } [numgrp $hrs]]" }
	if {$min} { append msg " $min [lindex {.   } [numgrp $min]]" }
	return [string trim $msg]
}


###   
bind msgm - * ::tpub::msg_pvt
proc ::tpub::msg_pvt {nick uhost hand text} { global botnick handlen
	if {[string range $text 0 0] == "!"} {
		return [msg_pub $nick $uhost $hand $nick "=$botnick [string range $text 1 end]"]
	}

	set cmd [string tolower [lindex [split $text] 0]]

	switch -exact -- $cmd {
		pass {
			if {[llength [split $text]] == 1} {
				putmsg $nick "?..   : pass "
				return
			}

			set pass [lindex [split $text] end]

			putlog "%$nick!$uhost% pass ..."

			if {$hand == "*"} {
				if {[string length $nick] > $handlen} {
					putmsg $nick "  .         $handlen  ,     ."
					return
				}
				if {[validuser $nick]} {
					putmsg $nick "     , ..   .     ,    \226  ,      ,  : ident ."
					return
				}
				if {![adduser $nick]} {
					putmsg $nick "- ...   !"
					return
				}
				chpass $nick $pass
				setuser $nick COMMENT "added by 3hauka.tcl"
				setuser $nick HOSTS [set host *![maskhost $uhost]]
				putmsg $nick ",   $host,     .       ,  \002ident \002."
				putlog "added user $nick with host $host"
			} elseif {[passwdok $hand -]} {
				chpass $hand $hand
				putmsg $nick ",   '$hand'."
				return
			} elseif {[passwdok $hand [lindex [split $text] 1]] && $pass != ""} {
				putlog "moo"
			}
		}
		ident {
			if {$hand != "*"} {
				putmsg $nick " -  \226 $hand ;)"
				return
			}
			if {[llength [split $text]] != 3} {
				putmsg $nick " .  : ident  ."
				return
			}
			if {![validuser [set tnick [lindex [split $text] end]]]} {
				putmsg $nick "     $tnick..."
				return
			}
			if {![passwdok $tnick [lindex [split $text] 1]]} {
				putmsg $nick "-      ..."
				putlog "%$nick!$uhost% IDENT $tnick failed!"
				return
			}
			putmsg $nick ",  [set host [maskhost $uhost]]    ."
			setuser $tnick HOSTS $host
			putlog "added mask $host to user $hand"
		}
		transfer {
			if {$hand == "*"} {
				putpvt $nick "    ."
				return
			}

			if {[passwdok $hand -]} {
				putpvt $nick "          .      : \002pass \002"
				return
			}

			set thand [lindex [split $text] 2]

			if {![validuser $thand]} {
				putpvt $nick "    \226 [lindex [split $text] 2]."
				return
			}

			if {![string compare $hand $thand]} {
				putpvt $nick ";)"
				return
			}

			if {![passwdok $hand [lindex [split $text] 1]]} {
				putpvt $nick " .    , \002transfer  \002"
				return
			}

			if {[catch {users check $thand} errmsg]} {
				putpvt $nick "- ,  !"
				putlog "3hauka.tcl: error in users.check: $errmsg"
				return
			}

			array set user1 [users get $hand]
			array set user2 [users get $thand]

			if {[catch {
				incr user2(score) $user1(score)
				incr user2(answers) $user1(answers)
				if {$user1(quick) < $user2(quick) && $user1(quick)} { set user2(quick) $user1(quick) }
				if {$user1(cont) > $user2(cont)} { set user2(cont) $user1(cont) }
				if {$user1(last) > $user2(last)} { set user2(last) $user1(last) }
				catch {
					if {$user2(email) == "" && $user1(email) != ""} { set user2(email) $user1(email) }
				}
				catch {
					if {$user2(memo) == "" && $user1(memo) != ""} { set user2(memo) $user1(memo) }
				}
			} errmsg]} {
				putpvt $nick "- ,  !"
				putlog "3hauka.tcl: error during transfer: $errmsg"
				return
			}

			users set [list $hand {}]
			users set [list $thand [array get user2]]

			putpvt $nick "      $thand."
			putlog "3hauka.tcl: transfer from $hand to $thand completed."
		}
		default {
			putlog "%$nick!$uhost% $cmd (what?)"
		}
	}
}


#       
proc ::tpub::filter {text} {
	regsub -all {[;:()\[\]!?.,/\\{}]} [tolower $text] "" text;
	return [string trim $text]
}

###    
bind pubm - * ::tpub::msg_pub
proc ::tpub::msg_pub {nick uhost hand chan text} { global botnick; variable conf
	#      .
	if {[string range $chan 0 0] == "#" && ![playground exists $chan]} {
		return
	}

	#   
	if {![playground exists $chan]} {
		return
	}

	#    (???)
	set text [encoding convertfrom [encoding system] $text]
	# [encoding system] == iso8859-1

	#   
	regsub -all {[\002\026\037]|\003\d{0,2}(,\d{0,2})?} $text "" text

	#  ,   
	if {![string compare -nocase "=$botnick" [lindex [split $text] 0]]} {
		set text "![join [lrange [split $text] 1 end]]"
	}

	#  
	if {[string range $text 0 0] != "!"} {
		variable game
		if {![info exists game] || $game(chan) != $chan || ![info exists game(question)]} { return }

		set game(cnt_lost) 0

		set text [filter $text]
		foreach answer [lindex $game(question) 1] {
			if {![string compare -nocase $text [filter [join $answer]]]} {
				#  
				if {![playground check $chan hide-answers]} {
					set correct "\226 \"\2$text\2\" \226 "
				} else {
					set correct ""
				}

				# ,   
				set time [expr [clock seconds] - $game(asktime)]

				if {$hand == "*"} {
					putmsg $chan "\00310, $nick!    ${correct}   \00304$time\00310 [lindex {.   } [numgrp $time]],       , ..     .\003"
					putmsg $nick " ,  ,     : \002pass \002 ( ).      ,    ,         (     , ?)."
					putmsg $nick "  ,  \226 ,       (  -, ),     .    \226    ,     ,  : \002ident \002"
					putmsg $nick "   ,   ,  : $conf(url-cmd)"
				} else {
					users check $hand

					array set usr [users get $hand]

					#   
					if {$game(lst_user) == $hand} {
						incr game(lst_cnt)
					} else {
						set game(lst_cnt) 1
						set game(lst_user) $hand
					}

					#    
					if {[info exists game(blitz)]} {
						incr game(points) $game(lst_cnt)
						incr game(points) -1
					}

					if {$game(lst_cnt) > 1 && $game(lst_cnt) > $usr(cont)} {
						set usr(cont) $game(lst_cnt)
						set msg_cont "$nick   $game(lst_cnt) [lindex {.   } [numgrp $game(lst_cnt)]] !"
					}

					#     
					if {$time < $usr(quick) || $usr(quick) < 1} {
						set usr(quick) $time
						set msg_time ",      ,"
					} else {
						set msg_time ""
					}

					#  
					incr usr(answers)
					incr usr(score) $game(points)

					#   
					incr game(cnt_won)

					putmsg $chan "\00310, $nick!    ${correct}   \00304$time\00310 [lindex {.   } [numgrp $time]]$msg_time   \00304 $game(points)\00310 [lindex {.   } [numgrp $game(points)]].\003"

					if {[info exists msg_cont]} {
						putmsg $chan "\00310$msg_cont\003"
					}

					set usr(last) [clock seconds]

					# store back
					users set [list $hand [array get usr]]
				}

				unset game(question)
				utimer 5 ::tpub::tmr_ask
				return 0
			}
		}

		#  
		if {[playground check $chan nightmare] && [llength [split $text]] == 1 && ![catch {array set usr [users get $hand]}]} {
			if {$game(points) == 2 && [info exists usr(score)] && $usr(score)} {
				incr usr(score) -1
				users set [list $hand [array get usr]]
			}
		}

		return
	}

	set cmd [string tolower [string range [lindex [split $text] 0] 1 end]]
	set text [join [lrange [split $text] 1 end]]

	#  
	if {[info commands ::tpub::cmd_$cmd] == ""} { return }

	#    (    )
	set user [expr {$hand == "*" ? "$nick!$uhost" : $hand}]

	# ...
	set status "OK"
	if {[catch {::tpub::cmd_$cmd $nick $chan $text} res]} {
		set status "ERROR"
		putlog "3hauka.tcl: error: $res"
		if {$hand != "*" && [string first o [chattr $hand]] != -1} {
			global errorCode errorInfo
			putnot $nick "error executing \002$cmd\002: $errorInfo"
		}
		set res 1
	} elseif {$res == "deny"} {
		set status "DENIED"
		putnot $nick "      \002$cmd\002,   ..."
	}

	if {$res != 0} {
		set msg "%$user% $cmd $text"
		if {$status != "OK"} { append msg " -- $status" }
		putlog $msg
	}
}


### help
proc ::tpub::cmd_help {nick chan text} { variable conf
	if {[info commands ::tpub::cmd_$text] != ""} {
		return [::tpub::cmd_$text $nick ? ?]
	}

	set cmds ""
	foreach item [lsort [info commands ::tpub::cmd_*]] {
		if {![string match "return \\\[cmd_* * * *\\\]" [string trim [info body $item]]]} { append cmds " [string range $item 12 end]" }
	}

	putpvt $nick "  ([llength $cmds]):$cmds.               \002!help \002,     $conf(url-cmd)"
}


### dump
proc ::tpub::cmd_dump {nick chan text} {
	if {$text == "" || "$chan $text" == "? ?"} { return [usage $nick dump text "     ."] }
	if {[string first m [chattr [nick2hand $nick]]] == -1} { return deny }
	putserv $text
}


###        
proc ::tpub::cmd_rehash {nick chan text} {
	if {"$chan $text" == "? ?"} { return [usage $nick rehash "" "     .      ."] }
	if {[string first m [chattr [nick2hand $nick]]] == -1} { return deny }

	if {$text == "questions"} {
		putnot $nick "rehashing questions..."
		question rehash
		return [cmd_next $nick $chan $text]
	}

	putnot $nick "rehashing..."
	rehash
}


###    
proc ::tpub::cmd_jump {nick chan text} {
	if {"$chan $text" == "? ?" || $text == ""} { return [usage $nick jump "" "    "] }
	if {[string first m [chattr [nick2hand $nick]]] == -1 } { return deny }

	putnot $nick "jumping..."
	jump [lindex [split $text] 0]
}


###  
proc ::tpub::cmd_restart {nick chan text} {
	if {"$chan $text" == "? ?"} { return [usage $nick restart "" "  ,       ."] }
	if {[string first n [chattr [nick2hand $nick]]] == -1} { return deny }
	restart
}


###  
proc ::tpub::cmd_append {nick chan text} {
	if {[llength $text] < 2} { return [usage $nick append "user_who user_to" "      ,  "] }

	if {[string first o [chattr [nick2hand $nick]]] == -1} { return deny }

	set h1 [lindex [split $text] 0]
	set h2 [lindex [split $text] 1]

	if {![validuser $h1]} {
		putnot $nick " $h1  ."
		return 0
	}

	if {$h1 == $h2} {
		putnot $nick "     ."
		return 0
	}

	array set user1 [users get $h1]
	array set user2 [users get $h2]

	# 
	catch { incr user2(score) $user1(score) }

	# 
	catch { incr user2(answers) $user1(answers) }

	#  
	catch { if {$user1(quick) < $user2(quick) && $user1(quick)} { set user2(quick) $user1(quick) } }

	#  ..
	catch { if {$user1(cont) > $user2(cont)} { set user2(cont) $user1(cont) } }

	# 
	catch { if {$user2(email) == "" && $user1(email) != ""} { set user2(email) $user1(email) } }

	# 
	catch { if {$user2(memo) == "" && $user1(memo) != ""} { set user2(memo) $user1(memo) } }

	#   
	catch { if {$user1(last) > $user2(last)} { set user2(last) $user1(last) } }

	users set [list $h1 {}]
	users set [list $h2 [array get user2]]

	putnot $nick "   $h1 ,       $h2."
}


###  
proc ::tpub::cmd_rename {nick chan text} {
	if {[llength $text] < 2} { return [usage $nick rename "nick_from nick_to" "  ."] }

	if {[string first o [chattr [nick2hand $nick]]] == -1} { return deny }

	set nick1 [lindex [split $text] 0]
	set nick2 [lindex [split $text] 1]

	if {![validuser $nick1]} {
		putnot $nick " $nick1  ."
		return 0
	}

	if {[validuser $nick2] && [string compare -nocase $nick1 $nick2]} {
		putnot $nick " $nick2  ."
		return 0
	}

	chhandle $nick1 $nick2

	putnot $nick " $nick1    $nick2."
}


###  
proc ::tpub::cmd_ {nick chan text} { return [cmd_op $nick $chan $text] }
proc ::tpub::cmd_op {nick chan text} { global botnick
	if {"$chan $text" == "? ?"} { return [usage $nick op "" "  ,     ."] }

	if {![validchan $chan]} {
		putnot $nick " !op      ,  ."
		return
	}

	if {![isop $botnick $chan]} {
		putnot $nick " ,      $chan,    .."
		return
	}

	if {[string first o [chattr [nick2hand $nick] $chan]] == -1} {
		return deny
	} else {
		pushmode $chan +o $nick
	}
}


###    
proc ::tpub::cmd_ {nick chan text} { return [cmd_kick $nick $chan $text] }
proc ::tpub::cmd_kick {nick chan text} { global botnick
	if {$text == "" || "$chan $text" == "? ?"} { return [usage $nick kick " ?...?" "   ,     (     )"] }

	if {![validchan $chan]} {
		putnot $nick " !kick      ,  ."
		return
	}

	if {[string first o [chattr [nick2hand $nick] $chan]] == -1} { return deny }

	if {![isop $botnick $chan]} {
		putnot $nick " ,       $chan,    ."
		return
	}

	set victim [string trim [string range $text 0 [string first \040 $text]]]
	if {![string compare -nocase $victim $botnick]} {
		putact $chan " ;)"
		return
	}

	if {![string compare -nocase $victim $nick]} {
		putact $chan "slaps $nick around a bit with a large trout!"
		return


	}

	if {[set reason [string trim [string range $text [string length $victim] end]]] == ""} {
		set reason "    ($nick)"
	}

	putserv "KICK $chan $victim :$reason"
}


###  
proc ::tpub::cmd_topic {nick chan text} {
	if {"$chan $text" == "? ?"} { return [usage $nick topic "?...?" "    ,      ."] }
	if {[string first o [chattr [nick2hand $nick]]] == -1} { return deny }
	putserv "TOPIC $chan :$text"
}


###    
proc ::tpub::cmd_ {nick chan text} { return [cmd_version $nick $chan $text] }
proc ::tpub::cmd_version {nick chan text} { variable game; variable conf
	global version tcl_patchLevel
	if {[info exists game] && $chan == $game(chan)} {
		putnot $nick "3hauka.tcl,  [getversion].         \037\00312http://quiz.forestnet.org/\003\037"
	} else {
		if {[info exists conf(url)]} {
			set homeinfo "       $conf(url)"
		} else { set homeinfo "" }
		putact $chan "\226  eggdrop v.[lindex $version 0] (tcl $tcl_patchLevel) + 3hauka.tcl  [getversion].$homeinfo"
	}
}


###  
proc ::tpub::cmd_list {nick chan text} {
	if {"$chan $text" == "? ?"} { return [usage $nick list "" "   ,    ."] }

	if {[string first o [chattr [nick2hand $nick] [expr {[validchan $chan] ? $chan : ""}]]] == -1} { return deny }

	if {![llength [set users [users list]]]} {
		putpvt $nick "   ."
	} else {
		set lst "([llength $users])"
		foreach item $users { append lst " [lindex $item 0]" }

		putpvt $nick "$lst."
	}
}


###    
proc ::tpub::cmd_ {nick chan text} { return [cmd_forget $nick $chan $text] }
proc ::tpub::cmd_forget {nick chan text} {
	if {$text == "" || "$chan $text" == "? ?"} { return [usage $nick forget "" "    ."] }

	if {[string first o [chattr [nick2hand $nick] [expr {[validchan $chan] ? $chan : ""}]]] == -1} { return deny }

	if {![validuser $text]} {
		putnot $nick "    $user ."
		return 0
	}

	users set [list $text {}]

	putact $chan " ,  -   $user"
}


###   
proc ::tpub::cmd_raw {nick chan text} {
	if {$text == "" || "$chan $text" == "? ?"} { return [usage $nick raw "" "     \"\" "] }

	if {[string first o [chattr [nick2hand $nick] [expr {[validchan $chan] ? $chan : ""}]]] == -1} { return deny }

	if {![validuser $text]} {
		putnot $nick " $text  ."
		return 0
	}

	if {[set xtra [users get $text]] == ""} {
		putnot $nick "   $text ."
		return 0
	}

	putpvt $nick "raw($text): [join $xtra]"
}


###  
proc ::tpub::cmd_set {nick chan text} {
	if {[llength [split $text]] < 2} { return [usage $nick set "  ??" "     (         \002raw\002),     "] }

	if {[string first o [chattr [nick2hand $nick] [expr {[validchan $chan] ? $chan : ""}]]] == -1} { return deny }

	if {![validuser [set user [lindex [split $text] 0]]]} {
		putnot $nick " \002$user\002  ."
		return 0
	}

	set para [lindex [split $text] 1]
	set data [lrange [split $text] 2 end]
	array set usr [users get $user]

	if {$data == ""} {
		if {[info exists usr($para)]} { unset usr($para) }
		putnot $nick "  $para   $user ."
	} else {
		set usr($para) $data
		putnot $nick "  $para   $user  : $data"
	}

	users set [list $user [array get usr]]
}


###  
proc ::tpub::cmd_memo {nick chan text} {
	if {$text == "" || "$chan $text" == "? ?"} { return [usage $nick memo "" "     ,    \002!\002,      ,   .      \"none\"."] }

	if {[set hand [nick2hand $nick]] == "*"} {
		putnot "       (     )."
		return 0
	}

	array set user [users get $hand]

	if {$text == ""} {
		if {[info exists user(memo)]} {
			putnot $nick " : $user(memo)"
		} else {
			putnot $nick "  ."
		}
		return 0
	} elseif {($text == "none") || ($text == "")} {
		if {[info exists user(memo)]} { unset user(memo) }
		putnot $nick " ."
	} else {
		set user(memo) [string range $text 0 199]
		putnot $nick "   : $user(memo)"
	}

	users set [list $hand [array get user]]
}


###  -
proc ::tpub::cmd_email {nick chan text} {
	if {$text == "" || "$chan $text" == "? ?"} { return [usage $nick email "" "  e-mail ,  , ,   "] }

	if {[set hand [nick2hand $nick]] == "*" || ![validuser $hand]} {
		putnot $nick "         (,      )."
		return 0
	}

	array set user [users get $hand]

	if {$text == ""} {
		if {![info exists user(email)]} {
			putnot $nick "    ."
		} else {
			putnot $nick "   : $user(email)"
		}
		return 0
	} elseif {$text == "none" || $text == ""} {
		if {[info exists user(email)]} { unset user(email) }
		putnot $nick "     ."
	} else {
		set user(email) [lindex [split $text] 0]
		putnot $nick "      $user(email)"
	}

	users set [list $hand [array get user]]
}


###  
proc ::tpub::cmd_ {nick chan text} { return [cmd_start $nick $chan $text] }
proc ::tpub::cmd_start {nick chan text} { variable game; variable conf
	if {"$chan $text" == "? ?"} { return [usage $nick start "" "    ,     ."] }

	if {[lsearch -exact [channel info $chan] "+bitch"] != -1 && [nick2hand $nick] == "*"} {
		return
	}

	if {[info exists game]} {
		if {$game(chan) == $chan} {
			putnot $nick "   :)"
		} else {
			putmsg $chan "$nick,      $game(chan),   ;)"
		}
		return
	}

	if {![set ::tpub::qcount] && ![question rehash]} {
		putmsg $chan "$nick,  ,       ,   ..."
		return
	}

	if {![playground exists $chan]} {
		putmsg $chan "$nick,      ."
		return
	}

	if {![checktime]} {
		putmsg $chan "$nick,       -     ."
		return
	}

	if {[set hand [nick2hand $nick]] == "*"} {
		global botnick
		putnot $nick "     .      \002/msg $botnick PASS \002,       ."
		return
	}

	set modes {blitz }

	if {$text != "" && [lsearch $modes $text] == -1} {
		putmsg $chan "$nick,      \226 $text, : [join $modes {, }]."
		return
	}

	if {[lsearch -exact {blitz } [lindex [split $text] 0]] != -1} {
		set game(blitz) 1
		set what "-"
	} else {
		set what ""
	}

	putmsg $chan "\00310    !   $what.     [set ::tpub::qcount] [lindex {.   } [numgrp [set ::tpub::qcount]]].\003"

	#     ,  
	users check [set hand [nick2hand $nick]]

	# 
	set game(chan) $chan			;# 
	set game(creator) $hand			;# ,   ( !stop)
	set game(started) [clock seconds]	;#   
	set game(cnt_lost) 0			;#   
	set game(cnt_won) 0			;#   
	set game(qid) 0				;#   
	set game(lst_user) ""			;#     
	set game(lst_cnt) 0			;#     

	#   
	utimer 5 ::tpub::tmr_ask
}


###    
proc ::tpub::cmd_ {nick chan text} { return [cmd_repeat $nick $chan $text] }
proc ::tpub::cmd_repeat {nick chan text} { variable game
	if {"$chan $text" == "? ?"} { return [usage $nick repeat "" "  ,   ."] }

	if {![info exists game]} {
		putmsg $chan "$nick,    -   \002!start\002"
		return 0
	}

	if {![info exists game(question)]} {
		putnot $nick "    ."
		return 0
	}

	putnot $nick " : [lindex $game(question) 0]"
	return 0
}


###  
proc ::tpub::cmd_ {nick chan text} { return [cmd_hint $nick $chan $text] }
proc ::tpub::cmd_hint {nick chan text} { variable game
	if {"$chan $text" == "? ?"} { return [usage $nick  "" "    ,    5 ,       20 .            (3 )."] }

	if {![info exists game] || $game(chan) != $chan || ![info exists game(question)]} {
		return 0
	}

	if {[info exists game(blitz)]} {
		return 0
	}

	#    
	foreach item [utimers] {
		if {[lrange [lindex $item 1] 0 1] == [list ::tpub::tmr_hint $game(qid)]} {
			return 0
		}
	}

	utimer 5 [list ::tpub::tmr_hint $game(qid)]
	return 0
}


###  
proc ::tpub::cmd_ {nick chan text} { return [cmd_next $nick $chan $text] }
proc ::tpub::cmd_next {nick chan text} { variable game
	if {"$chan $text" == "? ?"} { return [usage $nick next "" "  .    ,     ,  ."] }

	if {[string range $chan 0 0] != "#"} {
		putlog "3HAuKA: skip request from $nick ignored (not on channel)."
		return
	}

	# ,   .
	if {![info exists game]} {
		putmsg $chan "$nick,    -   \002!start\002"
		return
	}

	# 
	if {[playground check $chan no-skip]} {
		putlog "3HAuKA: skip request from $nick denied (playground configuration)."
		return
	}

	if {[string first o [chattr [nick2hand $nick $chan] [expr {[validchan $chan] ? $chan : ""}]]] == -1 && ![users isintop [set user [nick2hand $nick]]] && $game(creator) != [nick2hand $nick]} { return deny }

	set game(lst_user) ""
	set game(lst_cnt) 0
	set game(cnt_lost) 0

	if {![info exists game(question)]} { return 0 }
	tmr_next $game(qid)
}


###  
proc ::tpub::cmd_ {nick chan text} { return [cmd_stop $nick $chan $text] }
proc ::tpub::cmd_stop {nick chan text} { variable game
	if {"$chan $text" == "? ?"} { return [usage $nick stop "" "   .   ,     ,  ."] }

	if {![info exists game]} {
		putmsg $chan "$nick,    -   \002!start\002"
		return 0
	}

	if {[set user [nick2hand $nick]] != $game(creator) && [string first o [chattr [nick2hand $nick]]] == -1 && !([lsearch -exact [channel info $chan] "+bitch"] == -1 && [users isintop $user])} { return deny }

	set msg " ."

	if {$game(qid)} {
		append msg "  [lindex {.   } [numgrp $game(qid)]] $game(qid) [lindex {.   } [numgrp $game(qid)]]"
		if {$game(cnt_won)} { append msg ",    $game(cnt_won)    ." } else { append msg "." }

		append msg "   [game_time]."
	}

	putmsg $chan "\00304$msg\003"

	unset game
	users purge
}


###   
proc ::tpub::cmd_ {nick chan text} { return [cmd_top $nick $chan $text] }
proc ::tpub::cmd_top {nick chan text} { variable game
	if {"$chan $text" == "? ?"} { return [usage $nick top "??" "    ."] }

	#   
	variable conf
	if {[info exists game]} {
		if {[info exists game(topsent)] && [string first m [chattr [nick2hand $nick]]] == -1} {
			putlog "-- skipped --"
			return
		}
		set game(topsent) 1

		set msg ""
		set place 0
		foreach item [lrange [users list] 0 4] {
			append msg ", #[incr place] \226 [lindex $item 0] ([lindex $item 1])"
		}
		set url [expr {[info exists conf(url-top)] ? ".         \00312\037$conf(url-top)\037" : "."}]
		putmsg $chan "\0034  : [string range $msg 2 end]$url\003"
		return
	}

	if {![llength [set users [users list]]]} {
		putmsg $chan "$nick,   ."
		return 0
	}

	if {[catch {set count [expr int([lindex $text 0])]}]} { set count 5 }

	if {$count < 5} { set count 5 }
	if {$count > 9} { set count 10}

	putmsg $chan "\00304[lindex {     } [expr $count - 5]]    :\003"

	set idx 1
	foreach item $users {
		putmsg $chan "  \00304#[format %02d $idx]. [lindex $item 0] \226 \002[lindex $item 1]\002 [lindex {.   } [numgrp [lindex $item 1]]]\003"
		if {[incr idx] > $count} { break }
	}

	# $conf(url-top)
	putmsg $chan [expr {$idx > [llength $users] ? "\00304 , .  -    . ;)\003" : [expr {[info exists conf(url-top)] ? "\00304        \00312\037$conf(url-top)\003" : "\00304 , .     ... ;)\003"}] }]
}


###    
proc ::tpub::cmd_ {nick chan text} { return [cmd_won $nick $chan $text] }
proc ::tpub::cmd_won {nick chan text} { variable game
	if {"$chan $text" == "? ?"} { return [usage $nick won "??" "    ,    ."] }

	if {[string first "f" [chattr [nick2hand $nick] [expr {[validchan $chan] ? $chan : ""}]]] == -1 && [info exists game]} {
		if {[info exists game(wonsent)] && [lsearch $game(wonsent) $nick] != -1} {
			return
		}
		lappend game(wonsent) $nick
	}

	if {$text == ""} {
		set hand [nick2hand $nick]
	} elseif {![validuser [set hand [lindex [split $text] 0]]]} {
		set hand [nick2hand $hand]
	}

	if {[string is int $text] && $text != ""} {
		set list [users list]

		if {[expr int($text)] > [llength $list]} {
			putnot $nick "    [llength $list] [lindex {.   } [numgrp [llength $list]]]."
			return 0
		}

		set hand [lindex [lindex $list [expr int($text) - 1]] 0]
	}

	if {$hand == "" || $hand == "*"} {
		if {$text == ""} {
			putnot $nick "    ,  ? ;)"
		} else {
			putnot $nick " $text  ..."
		}
		return 0
	}

	array set user [set xtra [users get $hand]]

	set place -1
	if {![info exists list]} { set list [users list] }
	for {set i 0} {$i < [llength $list]} {incr i} {
		if {![string compare -nocase $hand [lindex [lindex $list $i] 0]]} {
			set place $i
			break
		}
	}
	incr place

	if {$xtra == ""} {
		putnot $nick "    $hand   ."
		return 0
	}

	set msg "  $hand  $user(answers) [lindex {.   } [numgrp $user(answers)]] ($user(score) [lindex {.   } [numgrp $user(score)]])"

	if {$user(cont) > 1} {
		append msg ",   \226 $user(cont) [lindex {.   } [numgrp $user(cont)]] "
	}

	append msg ",       $user(quick) [lindex {.   } [numgrp $user(quick)]]"

	if {[info exists user(memo)]} {
		append msg " (: $user(memo))"
	}

	if {$place} {
		set honor [expr {$place < 6 ? " " : ""}]
		append msg ".  $hand$honor $place- "
	}

	putmsg $chan "\00302$msg.\003"
}


###   
proc ::tpub::cmd_ {nick chan text} { return [cmd_stat $nick $chan $text] }
proc ::tpub::cmd_stat {nick chan text} { variable game
	if {"$chan $text" == "? ?"} { return [usage $nick stat "" "    ,     ."] }

	set usercount [llength [users list]]

	set msg "   [set ::tpub::qcount] [lindex {.   } [numgrp [set ::tpub::qcount]]],   [set count [question count]]   [lindex {.   } [numgrp $count]].  [lindex {.   } [numgrp $usercount]] $usercount [lindex {.   } [numgrp $usercount]]"

	if {[info exists game]} {
		if {$chan != $game(chan)} {
			append msg ".        $game(chan)"
		}
		append msg ".  [lindex {.   } [numgrp $game(qid)]] $game(qid) [lindex {.   } [numgrp $game(qid)]],    $game(cnt_won)    .    [game_time]"

		if {[info exists game(creator)]} {
			append msg ",      $game(creator)"
		}
	}

	putmsg $chan "\0034$msg.\003"
}


###   
bind join - * ::tpub::on_join
proc ::tpub::on_join {nick uhost handle chan} { variable game; variable conf
	if {![playground exists $chan] || $handle == "*"} { return }

	array set usr [users get $handle]

	if {[users isintop $handle]} {
		# if {$usr(score) > 100} { pushmode $chan +v $nick }
		if {[lsearch -exact [channel info $chan] "+greet"] != -1 && [info exists usr(memo)]} { putmsg $chan "\[\002$nick\002\] $usr(memo)" }
	}

	set not ""

	if {[info exists game]} {
		set not "   [game_time]"
		if {[info exists game(question)]} {
			append not ", : \"[lindex $game(question) 0]\""
		} else { append not "." }
	}

	if {$not != ""} { putnot $nick $not }
}


###   
### 123456
proc ::tpub::tmr_ask {} { variable game
	if {![info exists game] || [info exists game(question)]} { return }

	if {[info exists game(topsent)]} {
		unset game(topsent)
	}

	set game(question) [question get]
	set game(points) 3
	set game(asktime) [clock seconds]
	incr game(qid)

	putmsg $game(chan) "\00302[lindex $game(question) 0]\003"
	utimer [expr {[info exists game(blitz)] ? 20 : 60}] [list ::tpub::tmr_next $game(qid)]
}


###   
proc ::tpub::tmr_next {qid} { variable game
	if {![info exists game] || ![info exists game(question)] || $qid != $game(qid)} { return }

	#      ""
	set game(lst_user) ""
	set game(lst_cnt) 0

	if [info exists game(topsent)] { unset game(topsent) }
	if [info exists game(wonsent)] { unset game(wonsent) }

	# ,        
	set ct [checktime]

	#   
	if {!$ct || [incr game(cnt_lost)] == 3} {
		if {$game(cnt_lost) == 3} {
			set msg "        "
		} else {
			set msg "       "
		}

		if {$game(qid)} { append msg ".  [lindex {.   } [numgrp $game(qid)]] $game(qid) [lindex {.   } [numgrp $game(qid)]], " }

		if {$game(cnt_won)} {
			append msg "   $game(cnt_won)    ."
		} else {
			append msg "    ."
		}

		putmsg $game(chan) "\00304$msg   [game_time].\003"

		unset game
		return
	}

	putmsg $game(chan) "\00303    ,   5 .\003"
	unset game(question)
	utimer 5 ::tpub::tmr_ask
}


###  
proc ::tpub::tmr_hint {qid {len 2}} { variable game
	if {![info exists game] || ![info exists game(question)] || $qid != $game(qid)} { return }

	if {![incr game(points) -1]} {
		tmr_next $qid
		return
	}

	set answer [join [lindex [lindex $game(question) 1] 0]]

	if {$len >= [string length $answer]} { return }

	putmsg $game(chan) "\00312:\00302 [string range $answer 0 [expr $len - 1]]...?\003"

	utimer 20 [list ::tpub::tmr_hint $game(qid) [incr len]]
}


###  
namespace eval tpub {
	if {![set count [question count]]} { set count [question rehash] }
	putlog "3HAuKA-[getversion] here, $qcount [lindex {. question questions questions} [numgrp $qcount]], $count cached left."
}


# 
bind dcc m hint ::tpub::dcc_hint
proc ::tpub::dcc_hint {hand idx text} { variable game
	if {![info exists game]} {
		putidx $idx "Game is not started."
		return
	}

	if {![info exists game(question)]} {
		putidx $idx "Question isn't known yet."
		return
	}

	if {[info exists game(blitz)]} {
		putidx $idx "Uh-oh, it's a blitz!.."
		return
	}

	putidx $idx "Current answers: [string map { } [lindex $game(question) 1]]"
}


#   
bind dcc m won ::tpub::dcc_won
proc ::tpub::dcc_won {hand idx text} {
	if {![string length $text]} {
		putidx $idx "Usage: won <nick>"
		return
	}

	if {![validuser [set nick [lindex [split $text] 0]]]} {
		putidx $idx "Can't find anyone with such a nick."
		return
	}

	array set user [set xtra [users get $nick]]

	if {$xtra == ""} {
		putidx $idx "User $nick is not known to have played quiz."
	} elseif  {[catch {
		putlog "#$hand# won $nick"

		# count place number
		set place -1
		for {set list [users list]; set i 0} {$i < [llength $list]} {incr i} {
			if {![string compare -nocase $nick [lindex [lindex $list $i] 0]]} {
				set place $i
				break
			}
		}
		incr place

		set fame [expr {$place ? ", place: $place" : ""}]

		putidx $idx "User: $nick, score: $user(score), answers: $user(answers) (max: $user(cont)), speed: $user(quick) sec.$fame"

		if {[info exists user(memo)]} {
			putidx $idx "$nick has a memo: [string map { } $user(memo)]"
		}
	}]} { putidx $idx "Error displaying information for $nick, try: .tcl getuser $nick xtra quiz" }

	return 0
}


# 
bind dcc m score ::tpub::dcc_score
proc ::tpub::dcc_score {hand idx text} { variable game
	set fname [lindex [split $text] 1]

	switch -exact -- [lindex [split $text] 0] {
		"export" {
			if {$fname == ""} {
				putidx $idx "Usage: .score export filename"
				return 0
			}

			if {[catch {open $fname w} fileID]} {
				putidx $idx "Failed to open $fname for writing."
				return 0
			}

			set count 0
			foreach user [lsort [userlist]] {
				if {[set xtra [getuser $user xtra quiz]] == ""} { continue }
				puts $fileID "$user $xtra"
				incr count
			}

			close $fileID
			putidx $idx "$count record(s) exported to $fname"
			putcmdlog "#[idx2hand $idx]# score export $fname"
		}

		"import" {
			if {$fname == ""} {
				putidx $idx "Usage: .score import filename"
				return 0
			}

			if {[catch {open $fname r} fileID]} {
				putidx $idx "Failed to open $fname for reading."
				return 0
			}

			# remove old records
			foreach item [userlist] { setuser $item xtra quiz }

			set count 0
			set fails 0
			while {![eof $fileID]} {
				if {[set line [string trim [gets $fileID]]] == ""} { continue }
				if {![validuser [set user [lindex [split $line] 0]]]} {
					incr fails
				} else {
					setuser $user xtra quiz [join [lrange [split $line] 1 end]]
					incr count
				}
			}

			close $fileID
			putidx $idx "$count record(s) imported, $fails failed to import."
			putcmdlog "#[idx2hand $idx]# score import $fname"
		}

		default {
			putidx $idx "Usage: .score {import|export} filename"
			return 0
		}
	}
}


#   
if {[info exists ::tpub::conf(scoreboard)]} {
	bind time - "* * * * *" ::tpub::tmr_dump
} else {
	catch { unbind time - "* * * * *" ::tpub::tmr_dump }
}

proc ::tpub::tmr_dump {minute hour day month year} { variable conf
	if {[catch {open $conf(scoreboard) w} out]} { return }

	foreach user [lsort [userlist]] {
		if {[set xtra [getuser $user xtra quiz]] == ""} { continue }
		puts $out "$user $xtra"
	}

	close $out
}

# disable eggdrop's built-in help
# unbind msg - help *msg:help
